from ... import utils as _utils
from ... import preprocessing as _preprocessing

from ... import model as _model
from ... import onnx as _onnx


class MobileNetV2(_model.Model):
    '''
    MobileNetV2 classifier trained on ImageNet

    Model is taken from the `ONNX Model Zoo <https://github.com/onnx/models>`_.
    '''

    URL_MODEL = 'https://github.com/onnx/models/raw/master/vision/classification/mobilenet/model/mobilenetv2-7.onnx'
    URL_LABELS = 'https://github.com/onnx/models/raw/master/vision/classification/synset.txt'

    def __init__(self):
        '''
        '''
        super().__init__()

        # Load model
        self.model = _onnx.ONNXModel(self.URL_MODEL)

        # Load labels from text file
        labels_path = _utils.file_from_url(self.URL_LABELS, 'labels')
        self._labels = [
            ' '.join(line.strip().split()[1:])
            for line in open(labels_path)
        ]

    def forward(self, image, return_scores=False):
        '''
        :parameter image: np array.
            Input image
        :parameter return_scores: bool.
            If ``True``, return class scores.
        :return: ``str``: class label.
        '''
        scores = self.model(image)[0]
        label = self.labels[scores.argmax()]
        if return_scores:
            return label, scores
        else:
            return label

    def get_preprocessing(self):
        return _preprocessing.Preprocessing(
            resize=(224, 224),
            dtype='float32',
            divide_by_255=True,
            means=[0.485, 0.456, 0.406],
            stds=[0.229, 0.224, 0.225],
            padding=False,
            batch_dimension=True,
            channels_first=True,
        )

    @property
    def labels(self):
        '''
        :return: list of ImageNet classification labels
        '''
        return self._labels



from ... import utils as _utils
from ... import preprocessing as _preprocessing

from ... import model as _model
from ... import edgetpu as _edgetpu
from ... import output as _output


class SSDMobileNet(_model.Model):
    '''
    MobileNet V2 (or V1) SSD object detector trained on COCO dataset.

    Model is taken from the `google-coral repo <https://github.com/google-coral/edgetpu/tree/master/test_data>`_.

    Here is the `webcam demo <https://github.com/FastSense/nnio/tree/master/demos>`_ of an analogous model (:class:`nnio.zoo.onnx.detection.SSDMobileNetV1`) working.
    '''

    URL_CPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/ssd_mobilenet_{}_coco_quant_postprocess.tflite'
    URL_TPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/ssd_mobilenet_{}_coco_quant_postprocess_edgetpu.tflite'
    URL_LABELS = 'https://github.com/google-coral/edgetpu/raw/master/test_data/coco_labels.txt'

    def __init__(
        self,
        device='CPU',
        version='v2',
        threshold=0.5
    ):
        '''
        :parameter device: str.
            ``CPU`` by default.
            Set ``TPU`` or ``TPU:0`` to use the first EdgeTPU device.
            Set ``TPU:1`` to use the second EdgeTPU device etc.
        :parameter version: str.
            Either "v1" or "v2"
        :parameter threshold: float.
            Detection threshold. Affects the detector's sensitivity.
        '''
        super().__init__()

        self.threshold = threshold

        # Load model
        if device == 'CPU':
            model_path = self.URL_CPU.format(version)
        else:
            model_path = self.URL_TPU.format(version)
        self.model = _edgetpu.EdgeTPUModel(model_path, device)

        # Load labels from text file
        labels_path = _utils.file_from_url(self.URL_LABELS, 'labels_google')
        self._labels = {
            int(line.split()[0]): line.strip().split()[1]
            for line in open(labels_path)
        }

    def forward(self, image, return_info=False):
        '''
        :parameter image: np array.
            Input image
        :parameter return_info: bool.
            If ``True``, return inference time.
        :return: list of :class:`nnio.DetectionBox`
        '''
        out = self.model(image, return_info=return_info)
        if return_info:
            (boxes, classes, scores, _num_detections), info = out
        else:
            boxes, classes, scores, _num_detections = out
        # Parse output
        out_boxes = []
        for i in range(len(boxes[0])):
            if scores[0, i] < self.threshold:
                continue
            x_1, y_1, x_2, y_2 = boxes[0, i]
            label = self.labels[int(classes[0, i])]
            score = scores[0, i]
            out_boxes.append(
                _output.DetectionBox(x_1, y_1, x_2, y_2, label, score)
            )
        if return_info:
            return out_boxes, info
        else:
            return out_boxes

    def get_preprocessing(self):
        return _preprocessing.Preprocessing(
            resize=(300, 300),
            dtype='uint8',
            batch_dimension=True,
        )

    @property
    def labels(self):
        '''
        :return: list of COCO labels
        '''
        return self._labels



class SSDMobileNetFace(_model.Model):
    '''
    MobileNet V2 SSD face detector.

    Model is taken from the `google-coral repo <https://github.com/google-coral/edgetpu/tree/master/test_data>`_.
    '''

    URL_CPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/ssd_mobilenet_v2_face_quant_postprocess.tflite'
    URL_TPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/ssd_mobilenet_v2_face_quant_postprocess_edgetpu.tflite'

    def __init__(
        self,
        device='CPU',
        threshold=0.5
    ):
        '''
        :parameter device: str.
            ``CPU`` by default.
            Set ``TPU`` or ``TPU:0`` to use the first EdgeTPU device.
            Set ``TPU:1`` to use the second EdgeTPU device etc.
        :parameter threshold: float.
            Detection threshold. Affects the detector's sensitivity.
        '''
        super().__init__()

        self.threshold = threshold

        # Load model
        if device == 'CPU':
            model_path = self.URL_CPU
        else:
            model_path = self.URL_TPU
        self.model = _edgetpu.EdgeTPUModel(model_path, device)

    def forward(self, image, return_info=False):
        '''
        :parameter image: np array.
            Input image
        :parameter return_info: bool.
            If ``True``, return inference time.
        :return: list of :class:`nnio.DetectionBox`
        '''
        out = self.model(image, return_info=return_info)
        if return_info:
            (boxes, _, scores, _num_detections), info = out
        else:
            boxes, _, scores, _num_detections = out
        # Parse output
        out_boxes = []
        for i in range(len(boxes[0])):
            if scores[0, i] < self.threshold:
                continue
            x_1, y_1, x_2, y_2 = boxes[0, i]
            label = 'face'
            score = scores[0, i]
            out_boxes.append(
                _output.DetectionBox(x_1, y_1, x_2, y_2, label, score)
            )
        if return_info:
            return out_boxes, info
        else:
            return out_boxes

    def get_preprocessing(self):
        return _preprocessing.Preprocessing(
            resize=(320, 320),
            dtype='uint8',
            batch_dimension=True,
        )



from ... import utils as _utils
from ... import preprocessing as _preprocessing

from ... import model as _model
from ... import openvino as _openvino
from ... import output as _output


class SSDMobileNetV2(_model.Model):
    '''
    SSDMobileNetV2 object detection model trained on COCO dataset.

    Model is taken `from openvino <https://docs.openvinotoolkit.org/latest/omz_models_public_ssd_mobilenet_v2_coco_ssd_mobilenet_v2_coco.html>`_ and converted to openvino.

    Here is the `webcam demo <https://github.com/FastSense/nnio/tree/master/demos>`_ of an analogous model (:class:`nnio.zoo.onnx.detection.SSDMobileNetV1`) working.
    '''

    URL_MODEL_BIN = 'https://github.com/FastSense/nnio/raw/development/models/openvino/ssd_mobilenet_v2_coco/ssd_mobilenet_v2_coco_fp16.bin'
    URL_MODEL_XML = 'https://github.com/FastSense/nnio/raw/development/models/openvino/ssd_mobilenet_v2_coco/ssd_mobilenet_v2_coco_fp16.xml'
    URL_LABELS = 'https://github.com/amikelive/coco-labels/raw/master/coco-labels-paper.txt'

    def __init__(
        self,
        device='CPU',
        lite=True,
        threshold=0.5
    ):
        '''
        :parameter device: str.
            Choose Intel device:
            ``CPU``, ``GPU``, ``MYRIAD``.
            If there are multiple devices in your system, you can use indeces:
            ``MYRIAD:0`` but it is not recommended since Intel automatically chooses a free device.
        :parameter threshold: float.
            Detection threshold. It affects sensitivity of the detector.
        :parameter lite: bool.
            If True, use SSDLite version (idk exactly how it is lighter).
        '''
        super().__init__()

        self.threshold = threshold

        path_bin = self.URL_MODEL_BIN
        path_xml = self.URL_MODEL_XML
        if lite:
            path_bin = path_bin.replace('ssd', 'ssdlite')
            path_xml = path_xml.replace('ssd', 'ssdlite')

        # Load model
        self.model = _openvino.OpenVINOModel(path_bin, path_xml, device)

        # Load labels from text file
        labels_path = _utils.file_from_url(self.URL_LABELS, 'labels')
        self._labels = [
            line.strip()
            for line in open(labels_path)
        ]

    def forward(self, image, return_info=False):
        '''
        :parameter image: np array.
            Input image of a person.
        :parameter return_info: bool.
            If ``True``, return inference time.
        :return: list of :class:`nnio.DetectionBox`
        '''
        results = self.model(image, return_info=return_info)
        if return_info:
            results, info = results
        # Parse output
        out_boxes = []
        for res in results[0, 0]:
            _, label, score, y_1, x_1, y_2, x_2 = res
            if score < self.threshold:
                continue
            label = self.labels[int(label) - 1]
            out_boxes.append(
                _output.DetectionBox(x_1, y_1, x_2, y_2, label, score)
            )
        if return_info:
            return out_boxes, info
        else:
            return out_boxes

    def get_preprocessing(self):
        return _preprocessing.Preprocessing(
            resize=(300, 300),
            dtype='float32',
            channels_first=True,
            batch_dimension=True,
            bgr=True,
        )

    @property
    def labels(self):
        '''
        :return: list of COCO labels
        '''
        return self._labels



import time

from . import model as _model
from . import utils as _utils


class ONNXModel(_model.Model):
    '''
    This class is used with saved onnx models.

    Usage example::

        # Create model
        model = nnio.ONNXModel('path/to/model.onnx')
        # Create preprocessor
        preproc = nnio.Preprocessing(
            resize=(300, 300),
            dtype='uint8',
            batch_dimension=True,
            channels_first=True,
        )

        # Preprocess your numpy image
        image = preproc(image_rgb)

        # Make prediction
        class_scores = model(image)


    Using this class requires onnxruntime to be installed. See :ref:`installation`.
    '''
    def __init__(
        self,
        model_path: str,
    ):
        '''

        :parameter model_path: URL or path to the .onnx model
        '''
        super().__init__()
        # Download file from internet
        if _utils.is_url(model_path):
            model_path = _utils.file_from_url(model_path, 'models')
        # Load model and create inference session
        self.sess = self._make_interpreter(model_path)

    def forward(self, *inputs, return_info=False):
        assert len(inputs) == len(self.get_input_details())
        # List output names
        outputs = [
            info['name']
            for info in self.get_output_details()
        ]
        # Convert input to a dict
        inputs = {
            info['name']: inp
            for (info, inp) in zip(self.get_input_details(), inputs)
        }
        # Run network and measure time
        start = time.time()
        results = self.sess.run(outputs, inputs)
        end = time.time()
        # Process output a little
        if len(outputs) == 1:
            results = results[0]
        # Return results
        if return_info:
            info = {
                'invoke_time': end - start,
            }
            return results, info
        else:
            return results

    def get_input_details(self):
        return [
            {
                'name': info.name,
                'shape': info.shape,
                'dtype': info.type,
            }
            for info in self.sess.get_inputs()
        ]

    def get_output_details(self):
        return [
            {
                'name': info.name,
                'shape': info.shape,
                'dtype': info.type,
            }
            for info in self.sess.get_outputs()
        ]

    @staticmethod
    def _make_interpreter(model_path):
        'Load model and create inference session'
        import onnxruntime as rt
        sess = rt.InferenceSession(model_path)
        return sess



import urllib.request
import os
import pathlib
import getpass
import datetime
import numpy as np

from . import __version__

PACKAGE_NAME = 'nnio'

# Temperature logging flag
LOG_TEMPERATURE = False
temperature_files = {}

def is_url(s):
    '''
    Check if input string is url or not
    '''
    return s.startswith('http://') or s.startswith('https://')

def file_from_url(url, category='other'):
    '''
    Downloads file to "/home/$USER/.cache/nnio/" if it does not exist already.
    Returns path to the file
    '''
    # Get base path for file
    base_path = os.path.join(
        '/home',
        getpass.getuser(),
        '.cache',
        PACKAGE_NAME,
        '.'.join(__version__.split('.')[:2]),
        category,
    )
    # Create path if not exists
    if not os.path.exists(base_path):
        pathlib.Path(base_path).mkdir(parents=True, exist_ok=True)
    # Get file path
    file_name = url.split('/')[-1]
    file_path = os.path.join(
        base_path,
        file_name
    )
    # Download file from the url
    if not os.path.exists(file_path):
        print('Downloading file from: {}'.format(url))
        urllib.request.urlretrieve(url, file_path)
        print('Downloaded to: {}'.format(file_path))
    else:
        print('Using cached file: {}'.format(file_path))

    return file_path


# Flag setter
def enable_logging_temperature(enable=True):
    global LOG_TEMPERATURE
    LOG_TEMPERATURE = enable

def log_temperature(device, temperature):
    # Get path to file
    if device in temperature_files:
        file_path = temperature_files[device]
    else:
        # Get base path for file
        base_path = os.path.join(
            '/home',
            getpass.getuser(),
            '.telemetry',
        )
        # Create path if not exists
        if not os.path.exists(base_path):
            pathlib.Path(base_path).mkdir(parents=True, exist_ok=True)
        # Make file name
        dev_id = device.replace('MYRIAD', '')
        time_format = "vpu{}_%Y-%m-%d_%H-%M-%S".format(dev_id)
        file_name = datetime.datetime.now().strftime(time_format)
        file_path = os.path.join(base_path, file_name)
        # Remember file path
        temperature_files[device] = file_path
        # Write first line to this file
        with open(file_path, 'w') as f:
            f.write('time,vpu_temp\n')
            f.flush()

    val_time = datetime.datetime.now().isoformat(timespec='milliseconds')
    val_vpu_temp = int(temperature)
    with open(file_path, 'a') as f:
        f.write('{},{}\n'.format(val_time, val_vpu_temp))
        f.flush()


class HumanDataBase:
    def __init__(
        self,
        new_entity_threshold=0.25,
        merging_threshold=0.2
    ):
        self.vectors = {}
        self.counts = {}
        self.new_entity_threshold = new_entity_threshold
        self.merging_threshold = merging_threshold

    def find_closest(self, vec):
        keys = list(self.vectors.keys())
        vec = self.normalize(vec)
        distances = [
            self.distance(vec, self.vectors[key])
            for key in keys
        ]
        if len(distances) == 0:
            id_min = None
        else:
            id_min = np.argmin(distances)
        if id_min is None or distances[id_min] > self.new_entity_threshold:
            new_key = str(len(self.vectors))
            print('adding', new_key)
            if id_min is not None:
                print(distances[id_min])
            self.vectors[new_key] = vec
            self.counts[new_key] = 1
            return new_key
        else:
            key = keys[id_min]
            self.vectors[key] = self.vectors[key] * self.counts[key] + vec
            self.vectors[key] = self.normalize(self.vectors[key])
            self.counts[key] = self.counts[key] + 1

        return keys[id_min]

    def optimize(self):
        keys = list(self.vectors.keys())
        if len(keys) < 2:
            return
        # Find vectors which are too close
        distances = []
        for i in range(len(keys) - 1):
            for j in range(i + 1, len(keys)):
                dst = self.distance(
                    self.vectors[keys[i]],
                    self.vectors[keys[j]],
                )
                if dst < self.merging_threshold:
                    distances.append([i, j, dst])
        # Merge two clusters which are the closest
        if len(distances) > 0:
            idx = np.argmin(np.array(distances)[:, 2])
            i, j, dst = distances[idx]
            print('Merging {} with {}'.format(i, j))
            self.vectors[keys[i]] = self.vectors[keys[i]] + self.vectors[keys[j]]
            self.vectors[keys[i]] = self.normalize(self.vectors[keys[i]])
            self.counts[keys[i]] = 1
            del self.vectors[keys[j]]
            del self.counts[keys[j]]


    @staticmethod
    def normalize(vec):
        return vec / np.sqrt((vec**2).sum())
    
    @staticmethod
    def distance(vec1, vec2):
        return 1 - vec1 @ vec2



import cv2


class DetectionBox:
    def __init__(
        self,
        x_1,
        y_1,
        x_2,
        y_2,
        label=None,
        score=1.0,
    ):
        '''
        
        :parameter x_1: ``float`` in range ``[0, 1]``.
            Relative x coordinate of top-left corner.
        :parameter y_1: ``float`` in range ``[0, 1]``.
            Relative y coordinate of top-left corner.
        :parameter x_2: ``float`` in range ``[0, 1]``.
            Relative x coordinate of bottom-right corner.
        :parameter y_2: ``float`` in range ``[0, 1]``.
            Relative y coordinate of bottom-right corner.
        :parameter label: ``str`` or ``None``.
            Class label of the detected object.
        :parameter score: ``float``.
            Detection score
        '''
        self.x_1 = x_1
        self.x_2 = x_2
        self.y_1 = y_1
        self.y_2 = y_2
        self.label = label
        self.score = score

    def draw(
        self,
        image,
        color=(255,0,0),
        stroke_width=2,
        text_color=(255,0,0),
        text_width=2,
    ):
        '''
        Draws the detection box on an image

        :parameter image: numpy array.
        :parameter color: RGB color of the frame.
        :parameter stroke_width: boldness of the frame.
        :parameter text_color: RGB color of the text.
        :parameter text_width: boldness of the text.

        :return: Image with the box drawn on it.
        '''
        # Box corners:
        start_point = (
            int(image.shape[1] * self.y_1),
            int(image.shape[0] * self.x_1),
        )
        end_point = (
            int(image.shape[1] * self.y_2),
            int(image.shape[0] * self.x_2),
        )
        # Draw rectangle
        # pylint: disable=no-member
        image = cv2.rectangle(
            image,
            start_point,
            end_point,
            color, stroke_width)
        # Draw text
        if self.label is not None:
            # pylint: disable=no-member
            image = cv2.putText(
                image,
                self.label,
                (start_point[0], start_point[1] + 20 + stroke_width),
                cv2.FONT_HERSHEY_SIMPLEX,
                1, text_color, text_width, cv2.LINE_AA
            )
        return image

    def __str__(self):
        template = 'nnio.DetectionBox(x_1={}, y_1={}, x_2={}, y_2={}, label={}, score={})'
        s = template.format(
            self.x_1, self.y_1,
            self.x_2, self.y_2,
            self.label,
            self.score)
        return s



from ... import utils as _utils
from ... import preprocessing as _preprocessing

from ... import model as _model
from ... import onnx as _onnx


class OSNet(_model.Model):
    '''
    Omni-Scale Feature Network for Person Re-ID taken from `here <https://github.com/KaiyangZhou/deep-person-reid>`_ and converted to onnx.

    Here is the `webcam demo <https://github.com/FastSense/nnio/tree/master/demos>`_ of this model working.
    '''

    URL = 'https://github.com/FastSense/nnio/raw/master/models/person-reid/osnet_x1_0/osnet_x1_0_op10.onnx'

    def __init__(
        self,
    ):
        '''
        '''
        super().__init__()

        # Load model
        self.model = _onnx.ONNXModel(self.URL)

    def forward(self, image, return_info=False):
        '''
        :parameter image: np array.
            Input image of a person.
        :parameter return_info: bool.
            If ``True``, return inference time.
        :return: np.array of shape ``[512]`` - person appearence vector. You can compare them by cosine or Euclidian distance.
        '''
        out = self.model(image, return_info=return_info)
        if return_info:
            vector, info = out
        else:
            vector = out
        vector = vector[0]
        if return_info:
            return vector, info
        else:
            return vector

    def get_preprocessing(self):
        return _preprocessing.Preprocessing(
            resize=(128, 256),
            dtype='float32',
            divide_by_255=True,
            means=[0.485, 0.456, 0.406],
            stds=[0.229, 0.224, 0.225],
            channels_first=True,
            batch_dimension=True,
            padding=True,
        )
